
#ifndef DDH47_H
#define DDH47_H

#include "stdint.h"
#include "stdbool.h"
#include "math.h"

//
// Empfnger fr den Wettersender Pinneberg "DDH47" auf 147.3 MHz
//
// - Hardware
// - Software
//   - siehe Beispielcode "main.c"
// - Infos
//   - siehe "DCF77.h"
//
// Copyright: Public Domain
//
// C.B., 2021
//

static const char BMCodeTable[2*32] =
{
  //
  // "International telegraphy alphabet No. 2 (BaudotMurray code)"
  // (https://en.wikipedia.org/wiki/Baudot_code)
  //
  //..,  ....,  // dez bin    Letter Figure
  //------------------------------------------
  '\0',  '\0',  // 00  00000  Null
   'E',   '3',  // 01  00001  E      3
  '\n',  '\n',  // 02  00010  LF
   'A',   '-',  // 03  00011  A      -
   ' ',   ' ',  // 04  00100  Space
   'S',  '\'',  // 05  00101  S      '
   'I',   '8',  // 06  00110  I      8
   'U',   '7',  // 07  00111  U      7
  '\r',  '\r',  // 08  01000  CR
   'D',   '',  // 09  01001  D      "WRU?"
   'R',   '4',  // 10  01010  R      4
   // 'J',  '\7',  // 11  01011  J      Bell
   'J',   '',  // 11  01011  J      Bell
   'N',   ',',  // 12  01100  N      ,
   'F',   '!',  // 13  01101  F      !
   'C',   ':',  // 14  01110  C      :
   'K',   '(',  // 15  01111  K      (
   'T',   '5',  // 16  10000  T      5
   'Z',   '+',  // 17  10001  Z      +
   'L',   ')',  // 18  10010  L      )
   'W',   '2',  // 19  10011  W      2
   'H',   '',  // 20  10100  H      Pound
   'Y',   '6',  // 21  10101  Y      6
   'P',   '0',  // 22  10110  P      0
   'Q',   '1',  // 23  10111  Q      1
   'O',   '9',  // 24  11000  O      9
   'B',   '?',  // 25  11001  B      ?
   'G',   '&',  // 26  11010  G      &
  '\0',  '\0',  // 27  11011  Figure
   'M',   '.',  // 28  11100  M      .
   'X',   '/',  // 29  11101  X      /
   'V',   '=',  // 30  11110  V      =
  '\0',  '\0',  // 31  11111  Letter
};

enum TDDH47_Error
{
  _ErrOk                  =  0, // kein Fehler
  _ErrDCT                 =  1, // fr "ADCSamplesPerPeriod" gibt es keinen Source-Code
  _ErrCarrier_RawOverflow =  2, // CarrierRawCntPerSecond zu klein
  _ErrCarrier_RawCnt      =  3, // CarrierRawCntPerSecond ist nicht Vielfaches von CarrierFineCntPerSecond
  _ErrTimedAdc            =  4, // fCPU zu gross bzw. ADCSamplesPerPeriod zu klein bzw. TIM2 braucht Prescaler
  _ErrAdcSampleTime       =  5, // fCPU zu klein bzw. ADCSamplesPerPeriod zu gro
  _ErrAdcBufSize          =  6, // PeriodsPerSample ist nicht vielfaches von IRQPerSample
  _ErrAdcBufOverrun       =  7, // fCPU zu klein oder der IRQ war zu lange blockiert (wenn selten drfte Err zurckgesetzt werden)
};

enum TDDH47_Const
{
  fMHz                    = 64,    // STM32-CPU (zu klein->zu langsam fr ADC); Quarzgenauigkeit mit max. +-50 ppm erforderlich
  fCarrier                = 147300,// Senderkonstante
  dfCarrierMinMax         = 85,    // Senderkonstante -> fCarrier +- (dfCarrierMinMax/2)
  Baud                    = 50,    // Senderkonstante
  PeriodsPerSample        = 75,
  ADCRange                = 4096,  // STM32-Konstante
  ADCSamplesPerPeriod     = 8,     // untersttzt werden 3..12 Samples je Periode
  IRQPerSample            = 5,     // klein:viel RAM; gross:viel CPU-Last

  // berechnet
  Sample_MaxValue   = (int32_t)(128*ADCRange/2*PeriodsPerSample*ADCSamplesPerPeriod*4/M_PI), // Uss=VCC; Rechteck
  AdcCntPerSample   = PeriodsPerSample*ADCSamplesPerPeriod,
  AdcBufCnt         = 2*AdcCntPerSample/IRQPerSample,
  SMPPerSecond      = fCarrier/PeriodsPerSample,

  // Debug
  DbgSample_Cnt = 0*2048,
};

// hier sind die Ergebnisse (werden laufend vom IRQ aktualisiert)
typedef struct
{
  int8_t   Error;       // Fehler vom Typ TDDH47_Error (meist Konfigurationsfehler)
  int32_t  HalfBitCnt;  // laufender Zhler basierend auf STM32-Takt (nicht die Senderbitnummer)
  int32_t  SN_uV;       // Signalrauschabstand in V
  int32_t  SNR;         // Signalrauschverhltnis (ab 10 ausreichend; ab 20 sehr gut)
  int32_t  dfHSE_ppm;   // Idealwert fr ppm Parameter in DDH47_Init(); nur gltig wenn SNR>5
  char     TriteDecode; // vom einfachen Dekoder erkanntes Zeichen
  char     FancyDecode; // vom aufwndigen Dekoder erkanntes Zeichen (zeitverzgert)
} volatile TDDH47State;

// (... und ab hier sind nur interne Strukturen:)

typedef struct
{
  int16_t ValArr[AdcBufCnt]; // Puffer ADC-Werte
  uint8_t TimArr[AdcBufCnt]; // ADC-Sample-Zeitpunktabstnde
  int32_t w;
  int32_t n;
  int32_t a0;
  int32_t a1;
  int32_t d_ppm_L10;
} TTimedAdc;

enum TDDH47Control_Const
{
  DDH47_BitScale        = 16,
  DDH47_BitPhaseTau_s   = 16,
  DDH47_SyncCharCnt     = 10, // bercksichtigte Zeichenzahl fr Startbitsuche
  DDH47_HalbBitsPerChar = 15,
  DDH47_PhaseSumCnt     = DDH47_BitScale/2+1,
  DDH47_InfoSignalTau_s = 1,
  DDH47_InfoNoiseTau_s  = 1,
  DDH47_FSKAmpAvr_dn    = 3, // 0..3; in DDH47_BitScale Einheiten;

  DDH47_State0_Null           = 0,
  DDH47_State1_CalcChrQuality = 2,
  DDH47_State2_NextCharPhaseA = 3,
  DDH47_State3_NextCharPhaseB = 4,
  DDH47_State4_Info           = 5,

  // berechnet
  DDH47_AvrCnt = SMPPerSecond/Baud,
  DDH47_AmpCnt = SMPPerSecond/Baud,
  DDH47_ScaledBitCnt = DDH47_BitScale*DDH47_HalbBitsPerChar/2,
  DDH47_HBCnt = DDH47_SyncCharCnt*DDH47_HalbBitsPerChar,
};

typedef struct
{
  int32_t Amp;
  int32_t StartAmp;
  int8_t  Phase[2];
} THBItem;

typedef struct
{
  TTimedAdc TimedAdc;
  int32_t  PrevC;
  int32_t  PrevS;
  uint32_t dPhiFSK;
  int32_t  AvrNeg[2*DDH47_AvrCnt];
  int32_t  AvrPos[2*DDH47_AvrCnt];
  int32_t  AvrCur;
  int32_t  FSKAmpSum;
  int32_t  FSKAmpCnt;
  int32_t  SD_iQuarterBit;
  int32_t  SD_Byte;
  int32_t  SD_Balance;
  bool     SD_FigureShift;
  int32_t  InfoPrevAmp[2];
  int32_t  InfoSignal;
  int32_t  InfoSNR2;
  int32_t  InfoCurAmp;
  int32_t  Info_dfHSE;
  int32_t  Info_Cur_ppm;
  int32_t  AmpVal[DDH47_AmpCnt];
  int32_t  AmpCur;
  int32_t  State;
  int32_t  ScaledBitBalance;
  int32_t  ScaledBitCur;
  int32_t  ScaledBitHist[DDH47_ScaledBitCnt];
  int32_t  ScaledBitPhase;
  int32_t  ScaledBitPhaseMax;
  int32_t  ScaledBitPhasePos;
  int32_t  ScaledBitPhaseSum;
  int32_t  ScaledBitPhaseNext;
  int32_t  CharPhase;
  int32_t  CharPhaseNext;
  THBItem  HBArr[DDH47_HBCnt];
  int32_t  HBCur;
  int32_t  SyncMaxVal;
  int32_t  SyncMaxPos;
  bool     BMCodeFigureShift;
  int32_t  CurC;
  int32_t  CurS;
  uint32_t Rand;
  int32_t  AdcBufOfs;
  int32_t  IRQCur;
} TDDH47Control;

void DDH47_Init(TDDH47Control* Control, TDDH47State* State, int32_t ppm);
  // DDH47-Modul einschalten
  //  - HSE-Frequenz zuvor (selber) entsprechend "fMHz" (s.o.) einstellen
  //  - Parameter
  //    - Control : nur bereitstellen (es gibt nichts zu sehen oder zu ndern)
  //    - State   : Diese Struktur wird vom DDH47-Modul (TIM3_IRQHandler())
  //                laufend aktualisiert. Hier erhlt man smtliche Informationen.
  //                Die Inhalt darf nicht verndert werden.

void DDH47_Deinit(TDDH47Control* s);
  // DDH47-Modul abschalten

#endif
